static char rcsid[] = "$Id: transcript.c 224437 2021-07-21 16:53:46Z twu $";
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "transcript.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>		/* For qsort */
#include <ctype.h>		/* For isalpha */

#include "mem.h"
#include "sense.h"


/* Transcript_new */
#ifdef DEBUG0
#define debug0(x) x
#else
#define debug0(x)
#endif

/* Printing */
#ifdef DEBUG1
#define debug1(x) x
#else
#define debug1(x)
#endif

/* Concordance */
#ifdef DEBUG2
#define debug2(x) x
#else
#define debug2(x)
#endif

/* Transcript_trim */
#ifdef DEBUG5
#define debug5(x) x
#else
#define debug5(x)
#endif

/* Transcript tokens */
#ifdef DEBUG6
#define debug6(x) x
#else
#define debug6(x)
#endif


static int pairmax_transcriptome;
static int expected_pairlength;
static Outputtype_T output_type;
static Transcriptome_T transcriptome;

/* For debugging */
static Univ_IIT_T transcript_iit;


void
Transcript_setup (int pairmax_transcriptome_in, int expected_pairlength_in,
		  Outputtype_T output_type_in, Transcriptome_T transcriptome_in,
		  Univ_IIT_T transcript_iit_in) {

  pairmax_transcriptome = pairmax_transcriptome_in;
  expected_pairlength = expected_pairlength_in;
  output_type = output_type_in;
  transcriptome = transcriptome_in;
  
  /* For debugging */
  transcript_iit = transcript_iit_in;

  return;
}


#define T Transcript_T
struct T {
  int num;	      /* Index in transcriptome */
  int transcript_genestrand;	/* gplusp: > 0 if on plus strand and < 0 if on minus strand */
  int start;	      /* Starting coordinate on transcript, corresponding to querystart.  0-based */
  int end; /* Ending coordinate, corresponding to queryend.  If transcript_plusp is false, trend > trstart */
  int transcript_length;

  int trim_low;
  int trim_high;

  List_T tokens;

  Transcript_velocity_T velocity;
};



static List_T
push_token (List_T tokens, char *token, Listpool_T listpool) {
  char *copy;

  debug6(printf("Token is %s\n",token));
  copy = (char *) MALLOC((strlen(token)+1) * sizeof(char));
  strcpy(copy,token);
  return Listpool_push(tokens,listpool,(void *) copy);
}


/* Modified from Stage3end_remap_transcriptome */
static List_T
compute_tokens (int trnum, int transcript_genestrand, int trlow, int trhigh, int trlength,
		int trim_low, int trim_high, Listpool_T listpool) {
  Chrpos_T *exonstarts_array;
  int *exonbounds;
  int start_exoni, end_exoni, exoni, nexons;
  char token[6], firstchar, lastchar; /* For x%dx pattern, allows 3 digits for exoni */
  List_T tokens = NULL;
#ifdef DEBUG6
  char *label;
  bool allocp;
#endif

#ifdef DEBUG6
  label = Univ_IIT_label(transcript_iit,trnum,&allocp);
  printf("%s\n",label);
  if (allocp) {
    FREE(label);
  }
#endif

  nexons = Transcriptome_exons(&exonbounds,&exonstarts_array,transcriptome,trnum);
#ifdef DEBUG6
  for (exoni = 0; exoni < nexons; exoni++) {
    printf("%d\n",exonbounds[exoni]);
  }
#endif


  start_exoni = 0;
  while (start_exoni < nexons && exonbounds[start_exoni] <= trlow /* same as < trlow + 1*/) {
    start_exoni++;
  }
    
  end_exoni = start_exoni;
  while (end_exoni < nexons && exonbounds[end_exoni] < trhigh) {
    end_exoni++;
  }
  if (end_exoni == nexons) {
    end_exoni--;
  }

  debug6(printf("genestrand %d.  trlow %d, trhigh %d\n",transcript_genestrand,trlow,trhigh));
  debug6(printf("start_exon %d, end_exon %d\n",start_exoni,end_exoni));

  if (start_exoni == end_exoni) {
    /* Single exon */
    if (start_exoni == 0) {
      firstchar = trlow < 0 ? 'u' : '.';
    } else if (trlow != exonbounds[start_exoni - 1]) {
      firstchar = '.';
    } else if (transcript_genestrand > 0 && trim_low > 0) {
      firstchar = 's';
    } else if (transcript_genestrand < 0 && trim_high > 0) {
      firstchar = 's';
    } else {	    
      firstchar = '.';
    }
      
    if (end_exoni == nexons - 1) {
      lastchar = trhigh > trlength ? 'u' : '.';
    } else if (trhigh != exonbounds[end_exoni]) {
      lastchar = '.';
    } else if (transcript_genestrand > 0 && trim_high > 0) {
      lastchar = 's';
    } else if (transcript_genestrand < 0 && trim_low > 0) {
      lastchar = 's';
    } else {
      lastchar = '.';
    }
      
    sprintf(token,"%c%d%c",firstchar,start_exoni+1,lastchar);
    tokens = push_token(tokens,token,listpool);

  } else {
    /* First exon */
    if (start_exoni == 0) {
      firstchar = trlow < 0 ? 'u' : '.';
    } else if (trlow != exonbounds[start_exoni - 1]) {
      firstchar = '.';
    } else if (transcript_genestrand > 0 && trim_low > 0) {
      firstchar = 's';
    } else if (transcript_genestrand < 0 && trim_high > 0) {
      firstchar = 's';
    } else {	    
      firstchar = '.';
    }
    sprintf(token,"%c%ds",firstchar,start_exoni+1);
    tokens = push_token(tokens,token,listpool);

    for (exoni = start_exoni + 1; exoni < end_exoni; exoni++) {
      /* Middle exons */
      sprintf(token,"s%ds",exoni+1);
      tokens = push_token(tokens,token,listpool);
    }

    /* Last exon */
    if (end_exoni == nexons - 1) {
      lastchar = trhigh > trlength ? 'u' : '.';
    } else if (trhigh != exonbounds[end_exoni]) {
      lastchar = '.';
    } else if (transcript_genestrand > 0 && trim_high > 0) {
      lastchar = 's';
    } else if (transcript_genestrand < 0 && trim_low > 0) {
      lastchar = 's';
    } else {
      lastchar = '.';
    }

    sprintf(token,"s%d%c",end_exoni+1,lastchar);
    tokens = push_token(tokens,token,listpool);
  }

  debug6(printf("\n"));
  return List_reverse(tokens);
}  


/* Taken from Pair_tokens_free in pair.c */
static void
tokens_gc (List_T *tokens) {
  List_T p;
  char *token;

  for (p = *tokens; p != NULL; p = List_next(p)) {
    token = (char *) List_head(p);
    FREE(token);
  }
  /* List_free(&(*tokens)); -- allocated by Listpool_push */

  return;
}


/* Taken from Pair_tokens_copy in pair.c */
static List_T
tokens_copy (List_T old, Listpool_T listpool) {
  List_T new = NULL;
  char *new_token, *old_token;

  while (old != NULL) {
    old_token = (char *) List_head(old);
    new_token = (char *) MALLOC((strlen(old_token)+1) * sizeof(char));
    strcpy(new_token,old_token);
    new = Listpool_push(new,listpool,(void *) new_token);
    old = List_next(old);
  }

  return List_reverse(new);
}


#if 0
static void
tokens_print_stdout (List_T tokens) {
  List_T p;
  char *token;

  if (tokens != NULL) {
    token = (char *) List_head(tokens);
    printf("%s",token);
  
    for (p = List_next(tokens); p != NULL; p = List_next(p)) {
      token = (char *) List_head(p);
      printf("|%s",token);
    }
  }

  return;
}
#endif


/* Note that if both paired-ends are V_AMBIGUOUS, then the
   insertlength is also a clue, but the threshold should be decided
   and handled in post-processing */
static Transcript_velocity_T
tokens_velocity (List_T tokens) {
  bool validp = true, intronp = false, extendedp = false, splicedp = false;
  List_T p;
  char *token, first, last;
  int length;
  
  for (p = tokens; p != NULL; p = List_next(p)) {
    token = (char *) List_head(p);
    length = strlen(token);
    first = token[0];
    last = token[length-1];
    if (first == 'x' || last == 'x') {
      validp = false;
    }
    if (first == 'i' || last == 'i') {
      intronp = true;
    }
    if (first == 'u' || last == 'u') {
      extendedp = true;
    }
    if (first == 's' || last == 's') {
      splicedp = true;
    }
  }

  if (validp == false) {
    return V_INVALID;
  } else if (intronp == true) {
    return V_UNSPLICED;
  } else if (extendedp == true) {
    return V_EXTENDED;
  } else if (splicedp == true) {
    return V_SPLICED;
  } else {
    return V_AMBIGUOUS;
  }
}


int
Transcript_num (T this) {
  return this->num;
}

int
Transcript_start (T this) {
  return this->start;
}

int
Transcript_end (T this) {
  return this->end;
}

int
Transcript_length (T this) {
  return this->transcript_length;
}

int
Transcript_genestrand (T this) {
  return this->transcript_genestrand;
}

void
Transcript_free (T *old) {
  debug0(printf("Freeing %p\n",*old));
  tokens_gc(&(*old)->tokens);
  FREE(*old);

  return;
}

void
Transcript_gc (List_T *list) {
  List_T p;
  T this;

  for (p = *list; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    Transcript_free(&this);
  }
  /* List_free(&(*list)); -- allocated by Listpool_push */

  return;
}


/* querystart and queryend are for the read */
T
Transcript_new_need_tokens (int num, int transcript_genestrand,
			    int trstart, int trend, int trlength,
			    int trim_low, int trim_high, Listpool_T listpool) {
#ifdef DEBUG0
  char *label;
  bool allocp;
#endif

  T new = (T) MALLOC(sizeof(*new));

#ifdef DEBUG0
  label = Univ_IIT_label(transcript_iit,num,&allocp);
  printf("%p: Making transcript %s with trnum %d, genestrand %d, trstart %d, trend %d, trlength %d.  Need tokens\n",
	 new,label,num,transcript_genestrand,trstart,trend,trlength);
  if (allocp) {
    FREE(label);
  }
#endif

  new->num = num;
  new->transcript_genestrand = transcript_genestrand;
  new->start = trstart;
  new->end = trend;
  new->transcript_length = trlength;

  new->trim_low = trim_low;
  new->trim_high = trim_high;

  if (trstart < trend) {
    new->tokens = compute_tokens(num,transcript_genestrand,
				 /*trlow*/trstart,/*trhigh*/trend,trlength,
				 trim_low,trim_high,listpool);

  } else {
    new->tokens = compute_tokens(num,transcript_genestrand,
				 /*trlow*/trend,/*trhigh*/trstart,trlength,
				 trim_low,trim_high,listpool);
  }

  new->velocity = tokens_velocity(new->tokens);

  return new;
}


T
Transcript_new_with_tokens (int num, int transcript_genestrand,
			    int trstart, int trend, int trlength,
			    int querystart, int queryend, int querylength,
			    List_T tokens) {
  T new = (T) MALLOC(sizeof(*new));
#ifdef DEBUG0
  char *label;
  bool allocp;
#endif

#ifdef DEBUG0
  label = Univ_IIT_label(transcript_iit,num,&allocp);
  printf("%p: Making transcript %s with trnum %d, transcript_genestrand %d, start %d, end %d, trlength %d.  With tokens\n",
	 new,label,num,transcript_genestrand,trstart,trend,trlength);
  if (allocp) {
    FREE(label);
  }
#endif
  
  new->num = num;
  new->transcript_genestrand = transcript_genestrand;
  new->start = trstart;
  new->end = trend;
  new->transcript_length = trlength;

  if (transcript_genestrand > 0) {
    if (trstart < trend) {
      new->trim_low = querystart;
      new->trim_high = querylength - queryend;
    } else {
      new->trim_low = querylength - queryend;
      new->trim_high = querystart;
    }

  } else {
    if (trstart < trend) {
      new->trim_low = querylength - queryend;
      new->trim_high = querystart;
    } else {
      new->trim_low = querystart;
      new->trim_high = querylength - queryend;
    }
  }
  
  new->tokens = tokens;
  new->velocity = tokens_velocity(new->tokens);

  return new;
}


T
Transcript_copy (T old, Listpool_T listpool) {
  T new = (T) MALLOC(sizeof(*new));

  debug0(printf("Copying %p from %p\n",new,old));

  new->num = old->num;
  new->transcript_genestrand = old->transcript_genestrand;
  new->start = old->start;
  new->end = old->end;
  new->transcript_length = old->transcript_length;

  new->trim_low = old->trim_low;
  new->trim_high = old->trim_high;

  new->tokens = tokens_copy(old->tokens,listpool);
  new->velocity = old->velocity;

  return new;
}

List_T
Transcript_copy_list (List_T old, Listpool_T listpool) {
  List_T new = NULL, p;
  T this;

  for (p = old; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    new = Listpool_push(new,listpool,(void *) Transcript_copy(this,listpool));
  }
  return List_reverse(new);
}

void
Transcript_trim (T this, int trim_low_adj, int trim_high_adj, Listpool_T listpool) {
#ifdef DEBUG5
  char *label;
  bool allocp;
#endif

#ifdef DEBUG5
  label = Univ_IIT_label(transcript_iit,this->num,&allocp);
  printf("Transcript_trim %s: trim_low_adj %d, trim_high_adj %d\n",
	 label,trim_low_adj,trim_high_adj);
  if (allocp) {
    FREE(label);
  }
#endif

  if (trim_low_adj == 0 && trim_high_adj == 0) {
    return;
  }

  tokens_gc(&this->tokens);

  this->trim_low += trim_low_adj;
  this->trim_high += trim_high_adj;

  if (this->transcript_genestrand > 0) {
    if (this->start < this->end) {
      if (trim_low_adj != 0) {
	this->start += trim_low_adj;
      }
      if (trim_high_adj != 0) {
	this->end -= trim_high_adj;
      }
      this->tokens = compute_tokens(this->num,this->transcript_genestrand,
				    /*trlow*/this->start,/*trhigh*/this->end,this->transcript_length,
				    this->trim_low,this->trim_high,listpool);

    } else {
      if (trim_low_adj != 0) {
	this->end += trim_low_adj;
      }
      if (trim_high_adj != 0) {
	this->start -= trim_high_adj;
      }
      this->tokens = compute_tokens(this->num,this->transcript_genestrand,
				    /*trlow*/this->end,/*trhigh*/this->start,this->transcript_length,
				    this->trim_low,this->trim_high,listpool);
    }

  } else {
    if (this->start < this->end) {
      if (trim_low_adj != 0) {
	this->end -= trim_low_adj;
      }
      if (trim_high_adj != 0) {
	this->start += trim_high_adj;
      }
      this->tokens = compute_tokens(this->num,this->transcript_genestrand,
				    /*trlow*/this->start,/*trhigh*/this->end,this->transcript_length,
				    this->trim_low,this->trim_high,listpool);

    } else {
      if (trim_low_adj != 0) {
	this->start -= trim_low_adj;
      }
      if (trim_high_adj != 0) {
	this->end += trim_high_adj;
      }
      this->tokens = compute_tokens(this->num,this->transcript_genestrand,
				    /*trlow*/this->end,/*trhigh*/this->start,this->transcript_length,
				    this->trim_low,this->trim_high,listpool);
    }
  }

  this->velocity = tokens_velocity(this->tokens);

  return;
}


bool
Transcript_extend_utr5p (List_T transcripts) {
  bool extend5p = true;
  List_T p;
  T this;

  for (p = transcripts; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    if (this->start < this->end) {
      if (this->start != 0) {
	extend5p = false;
      }
    } else {
      if (this->end != 0) {
	extend5p = false;
      }
    }
  }

  return extend5p;
}


bool
Transcript_extend_utr3p (List_T transcripts) {
  bool extend3p = true;
  List_T p;
  T this;

  for (p = transcripts; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    if (this->start < this->end) {
      if (this->end != this->transcript_length) {
	extend3p = false;
      }
    } else {
      if (this->start != this->transcript_length) {
	extend3p = false;
      }
    }
  }

  return extend3p;
}



#if 0
static bool
Transcript_equal (T x, T y) {
  if (x->num == y->num &&
      x->start == y->start &&
      x->end == y->end) {
    return true;
  } else {
    return false;
  }
}
#endif

bool
Transcript_in_list_p (T x, List_T list) {
  List_T p;
  T y;

  for (p = list; p != NULL; p = List_next(p)) {
    y = (T) List_head(p);
    if (x->num == y->num &&
	x->start == y->start &&
	x->end == y->end) {
      return true;
    }
  }
  return false;
}


#ifdef DEBUG1
void
Transcript_print_nums (List_T list) {
  List_T p;
  T this;

  printf(" Trnums:");
  for (p = list; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    printf(" %d:%d",this->num,this->start);
  }
  return;
}
#endif

#ifdef DEBUG1
void
Transcript_print_list_debug (List_T list) {
  List_T p;
  T this;

  printf(" Trnums:");
  for (p = list; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    printf(" %d",this->num);
  }
  printf("\n");

  printf(" Trstarts:");
  for (p = list; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    printf(" %d",this->start);
  }
  printf("\n");

  printf(" Trends:");
  for (p = list; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    printf(" %d",this->end);
  }
  printf("\n");

  return;
}
#endif

#if 0
int
Transcript_list_sensedir (List_T transcripts) {
  int sensedir = SENSE_NULL;
  bool tplusp, conflictp = false;
  List_T p;
  T transcript;
  
  for (p = transcripts; p != NULL; p = List_next(p)) {
    transcript = (T) List_head(p);
    tplusp = transcript->start < transcript->end ? true : false;

    if (tplusp == true) {
      if (sensedir == SENSE_NULL) {
	sensedir = SENSE_FORWARD;
      } else if (sensedir == SENSE_ANTI) {
	conflictp = true;
      }

    } else {
      if (sensedir == SENSE_NULL) {
	sensedir = SENSE_ANTI;
      } else if (sensedir == SENSE_FORWARD) {
	conflictp = true;
      }
    }
  }

  if (conflictp == true) {
    return SENSE_NULL;
  } else {
    return sensedir;
  }
}
#endif


/* Puts V_SPLICED transcripts first and V_INVALID transcripts last */
static int
Transcript_velocity_cmp (const void *a, const void *b) {
  T x = * (T *) a;
  T y = * (T *) b;

  if (x->velocity > y->velocity) {
    return -1;
  } else if (y->velocity > x->velocity) {
    return +1;
  } else if (x->num < y->num) {
    return -1;
  } else if (y->num < x->num) {
    return +1;
  } else if (x->start < y->start) {
    return -1;
  } else if (y->start < x->start) {
    return +1;
  } else if (x->end < y->end) {
    return -1;
  } else if (y->end < x->end) {
    return +1;
  } else {
    return 0;
  }
}


static int
Transcript_cmp (const void *a, const void *b) {
  T x = * (T *) a;
  T y = * (T *) b;

  if (x->num < y->num) {
    return -1;
  } else if (y->num < x->num) {
    return +1;
  } else if (x->start < y->start) {
    return -1;
  } else if (y->start < x->start) {
    return +1;
  } else if (x->end < y->end) {
    return -1;
  } else if (y->end < x->end) {
    return +1;
  } else {
    return 0;
  }
}


static int
velocity_accept_cmp (Transcript_velocity_T velocity1, Transcript_velocity_T velocity2) {
  if (velocity1 == V_SPLICED && velocity2 == V_AMBIGUOUS) {
    return -1;
  } else if (velocity2 == V_SPLICED && velocity1 == V_AMBIGUOUS) {
    return +1;
  } else if (velocity1 < velocity2) {
    return -1;
  } else if (velocity2 < velocity1) {
    return +1;
  } else {
    return 0;
  }
}


static bool
concordant_p (int *desired_genestrand, Transcript_velocity_T *accept_velocity,
	      T *array5, T *array3, int ntranscripts5, int ntranscripts3) {
  bool foundp = false, conflictp = false;
  int genestrand;
  int i, j, k, l;
  int a, b;
  int trnum5, trnum3;
  int start5, end5, end3;

#ifdef DEBUG2
  char *label;
  bool allocp;
#endif


  *desired_genestrand = 0;
  *accept_velocity = V_AMBIGUOUS;

  debug2(printf("Entered concordant_p with %d and %d transcripts\n",ntranscripts5,ntranscripts3));

  /* Re-sort beginning of arrays ignoring velocity */
  qsort(array5,ntranscripts5,sizeof(T),Transcript_cmp);
  qsort(array3,ntranscripts3,sizeof(T),Transcript_cmp);

#ifdef DEBUG2
  for (i = 0; i < ntranscripts5; i++) {
    label = Univ_IIT_label(transcript_iit,array5[i]->num,&allocp);
    printf("5' Transcript %s, number %d\n",label,array5[i]->num);
  }
  for (i = 0; i < ntranscripts3; i++) {
    label = Univ_IIT_label(transcript_iit,array3[i]->num,&allocp);
    printf("3' Transcript %s, number %d\n",label,array3[i]->num);
  }
#endif


  i = k = 0;
  while (i < ntranscripts5 && k < ntranscripts3) {
    trnum5 = array5[i]->num;
    j = i+1;
    while (j < ntranscripts5 && array5[j]->num == trnum5) {
      j++;
    }

    trnum3 = array3[k]->num;
    l = k+1;
    while (l < ntranscripts3 && array3[l]->num == trnum3) {
      l++;
    }
    
    if (trnum5 < trnum3) {
      i = j;
      
    } else if (trnum3 < trnum5) {
      k = l;
      
    } else {
      for (a = i; a < j; a++) {
	start5 = array5[a]->start;
	end5 = array5[a]->end;
	
	if (start5 < end5) {
	  /* Plus on transcript */
	  for (b = k; b < l; b++) {
	    end3 = array3[b]->end;
#ifdef DEBUG2
	    label = Univ_IIT_label(transcript_iit,array5[a]->num,&allocp);
	    printf("plus: checking for concordance on %s, trnum %d: start5 %d, end3 %d, pairmax %d\n",
		   label,trnum5,start5,end3,pairmax_transcriptome);
#endif
	    if (start5 < end3 && end3 <= start5 + pairmax_transcriptome) {
	      if ((genestrand = array3[b]->transcript_genestrand) > 0 && *desired_genestrand < 0) {
		conflictp = true;
	      } else if (genestrand < 0 && *desired_genestrand > 0) {
		conflictp = true;
	      } else {
		*desired_genestrand = genestrand;
	      }

	      /* printf("Comparing %d and %d\n",*accept_velocity,array5[a]->velocity); */
	      if (velocity_accept_cmp(*accept_velocity,array5[a]->velocity) > 0) {
		*accept_velocity = array5[a]->velocity;
		/* printf("accept_velocity is now %d\n",*accept_velocity); */
	      }
	      /* printf("Comparing %d and %d\n",*accept_velocity,array3[b]->velocity); */
	      if (velocity_accept_cmp(*accept_velocity,array3[b]->velocity) > 0) {
		*accept_velocity = array3[b]->velocity;
		/* printf("accept_velocity is now %d\n",*accept_velocity); */
	      }

	      foundp = true;
	    }
	  }
	  
	} else {
	  /* Minus on transcript */
	  for (b = k; b < l; b++) {
	    end3 = array3[b]->end;
#ifdef DEBUG2
	    label = Univ_IIT_label(transcript_iit,array3[b]->num,&allocp);
	    printf("minus: checking for concordance on %s, trnum %d: end3 %d, start5 %d, pairmax %d\n",
		   label,trnum5,end3,start5,pairmax_transcriptome);
#endif
	    if (end3 < start5 && start5 <= end3 + pairmax_transcriptome) {
	      if ((genestrand = array3[b]->transcript_genestrand) > 0 && *desired_genestrand < 0) {
		conflictp = true;
	      } else if (genestrand < 0 && *desired_genestrand > 0) {
		conflictp = true;
	      } else {
		*desired_genestrand = genestrand;
	      }

	      /* printf("Comparing %d and %d\n",*accept_velocity,array5[a]->velocity); */
	      if (velocity_accept_cmp(*accept_velocity,array5[a]->velocity) > 0) {
		*accept_velocity = array5[a]->velocity;
		/* printf("accept_velocity is now %d\n",*accept_velocity); */
	      }
	      /* printf("Comparing %d and %d\n",*accept_velocity,array3[b]->velocity); */
	      if (velocity_accept_cmp(*accept_velocity,array3[b]->velocity) > 0) {
		*accept_velocity = array3[b]->velocity;
		/* printf("accept_velocity is now %d\n",*accept_velocity); */
	      }

	      foundp = true;
	    }
	  }
	}
      }
      
      i = j;
      k = l;
    }
  }

  if (conflictp == true) {
    *desired_genestrand = 0;
  }

  debug2(printf("Returning foundp %d\n\n",foundp));
  return foundp;
}


bool
Transcript_concordant_p (int *desired_genestrand,
			 Transcript_velocity_T *search_velocity,
			 Transcript_velocity_T *accept_velocity,
			 List_T transcripts5, List_T transcripts3) {
  T *array5, *array3, transcript;
  List_T p;
  int n5, n3, nspliced5, nspliced3, nambiguous5, nambiguous3, nextended5, nextended3,
    nunspliced5, nunspliced3, ntranscripts5, ntranscripts3;
#ifdef DEBUG2
  char *label;
  bool allocp;
#endif
  
  debug2(printf("Entered Transcript_concordant_p with %d 5' and %d 3' transcripts\n",
		List_length(transcripts5),List_length(transcripts3)));

  if (transcripts5 != NULL && transcripts3 != NULL) {
    nspliced5 = nambiguous5 = nextended5 = nunspliced5 = ntranscripts5 = 0;
    for (p = transcripts5; p != NULL; p = List_next(p)) {
      transcript = (T) List_head(p);
#ifdef DEBUG2
      label = Univ_IIT_label(transcript_iit,transcript->num,&allocp);
      printf("5' Transcript %s (%p): %d..%d, velocity %d: ",
	     label,transcript,transcript->start,transcript->end,transcript->velocity);
      tokens_print_stdout(transcript->tokens);
      printf("\n");
#endif

      if (transcript->velocity == V_SPLICED) {
	nspliced5++;
      } else if (transcript->velocity == V_AMBIGUOUS) {
	nambiguous5++;
      } else if (transcript->velocity == V_EXTENDED) {
	nextended5++;
      } else if (transcript->velocity == V_UNSPLICED) {
	nunspliced5++;
      }
      ntranscripts5++;
    }
    debug2(printf("Have %d spliced5, %d ambiguous5, %d extended5, %d unspliced5\n",
		  nspliced5,nambiguous5,nextended5,nunspliced5));


    nspliced3 = nambiguous3 = nextended3 = nunspliced3 = ntranscripts3 = 0;
    for (p = transcripts3; p != NULL; p = List_next(p)) {
      transcript = (T) List_head(p);
#ifdef DEBUG2
      label = Univ_IIT_label(transcript_iit,transcript->num,&allocp);
      printf("3' Transcript %s (%p): %d..%d, velocity %d: ",
	     label,transcript,transcript->start,transcript->end,transcript->velocity);
      tokens_print_stdout(transcript->tokens);
      printf("\n");
#endif

      if (transcript->velocity == V_SPLICED) {
	nspliced3++;
      } else if (transcript->velocity == V_AMBIGUOUS) {
	nambiguous3++;
      } else if (transcript->velocity == V_EXTENDED) {
	nextended3++;
      } else if (transcript->velocity == V_UNSPLICED) {
	nunspliced3++;
      }
      ntranscripts3++;
    }
    debug2(printf("Have %d spliced3, %d ambiguous3, %d extended3, %d unspliced3\n",
		  nspliced3,nambiguous3,nextended3,nunspliced3));


    array5 = (T *) MALLOC(ntranscripts5*sizeof(T));
    array3 = (T *) MALLOC(ntranscripts3*sizeof(T));
    List_fill_array((void **) array5,transcripts5);
    List_fill_array((void **) array3,transcripts3);
    qsort(array5,ntranscripts5,sizeof(T),Transcript_velocity_cmp);
    qsort(array3,ntranscripts3,sizeof(T),Transcript_velocity_cmp);

    n5 = n3 = 0;
    if (nspliced5 > 0 || nspliced3 > 0) {
      n5 += nspliced5; n3 += nspliced3;
      /* Consider V_SPLICED */
      if (concordant_p(&(*desired_genestrand),&(*accept_velocity),array5,array3,n5,n3) == true) {
	FREE(array3);
	FREE(array5);
	*search_velocity = V_SPLICED;
	debug2(printf("%d %d Returning velocity %d..%d\n",n5,n3,*search_velocity,*accept_velocity));
	return true;
      }
    }

    if (nambiguous5 > 0 || nambiguous3 > 0) {
      n5 += nambiguous5; n3 += nambiguous3;
      /* Consider V_AMBIGUOUS */
      if (concordant_p(&(*desired_genestrand),&(*accept_velocity),array5,array3,n5,n3) == true) {
	FREE(array3);
	FREE(array5);
	*search_velocity = V_AMBIGUOUS;
	debug2(printf("%d %d Returning velocity %d..%d\n",n5,n3,*search_velocity,*accept_velocity));
	return true;
      }
    }

    if (nextended5 > 0 || nextended3 > 0) {
      n5 += nextended5; n3 += nextended3;
      /* Consider V_EXTENDED */
      if (concordant_p(&(*desired_genestrand),&(*accept_velocity),array5,array3,n5,n3) == true) {
	FREE(array3);
	FREE(array5);
	*search_velocity = V_EXTENDED;
	debug2(printf("%d %d Returning velocity %d..%d\n",n5,n3,*search_velocity,*accept_velocity));
	return true;
      }
    }

    if (nunspliced5 > 0 || nunspliced3 > 0) {
      n5 += nunspliced5; n3 += nunspliced3;
      /* Consider V_UNSPLICED */
      if (concordant_p(&(*desired_genestrand),&(*accept_velocity),array5,array3,n5,n3) == true) {
	FREE(array3);
	FREE(array5);
	*search_velocity = V_UNSPLICED;
	debug2(printf("%d %d Returning velocity %d..%d\n",n5,n3,*search_velocity,*accept_velocity));
	return true;
      }
    }

    if (concordant_p(&(*desired_genestrand),&(*accept_velocity),array5,array3,ntranscripts5,ntranscripts3) == true) {
      FREE(array3);
      FREE(array5);
      *search_velocity = V_INVALID;
      debug2(printf("%d %d Returning velocity %d..%d\n",n5,n3,*search_velocity,*accept_velocity));
      return true;
    }

    FREE(array3);
    FREE(array5);
  }
  
  debug2(printf("Returning false\n"));
  return false;
}


List_T
Transcript_get_concordant (int desired_genestrand,
			   Transcript_velocity_T search_velocity,
			   Transcript_velocity_T accept_velocity,
			   List_T transcripts5, List_T transcripts3,
			   Listpool_T listpool, bool first_read_p) {
  List_T concordant = NULL, p;
  Transcript_velocity_T velocity;
  
  T *array5, *array3, transcript;
  int n5, n3, nspliced5, nspliced3, nambiguous5, nambiguous3, nextended5, nextended3,
    nunspliced5, nunspliced3, ntranscripts5, ntranscripts3;

  int i, j, k, l;
  int besta, bestb, a, b;
  int trnum5, trnum3;
  int start5, end5, end3;
  int best_absdifflength, absdifflength, insertlength;


  debug2(printf("Entered Transcript_get_concordant with velocity %d..%d, first_read_p %d\n",
		search_velocity,accept_velocity,first_read_p));

  nspliced5 = nambiguous5 = nextended5 = nunspliced5 = ntranscripts5 = 0;
  for (p = transcripts5; p != NULL; p = List_next(p)) {
    transcript = (T) List_head(p);
    if (transcript->velocity == V_SPLICED) {
      nspliced5++;
    } else if (transcript->velocity == V_AMBIGUOUS) {
      nambiguous5++;
    } else if (transcript->velocity == V_EXTENDED) {
      nextended5++;
    } else if (transcript->velocity == V_UNSPLICED) {
      nunspliced5++;
    }
    ntranscripts5++;
  }

  nspliced3 = nambiguous3 = nextended3 = nunspliced3 = ntranscripts3 = 0;
  for (p = transcripts3; p != NULL; p = List_next(p)) {
    transcript = (T) List_head(p);
    if (transcript->velocity == V_SPLICED) {
      nspliced3++;
    } else if (transcript->velocity == V_AMBIGUOUS) {
      nambiguous3++;
    } else if (transcript->velocity == V_EXTENDED) {
      nextended3++;
    } else if (transcript->velocity == V_UNSPLICED) {
      nunspliced3++;
    }
    ntranscripts3++;
  }


  array5 = (T *) MALLOC(ntranscripts5*sizeof(T));
  array3 = (T *) MALLOC(ntranscripts3*sizeof(T));
  List_fill_array((void **) array5,transcripts5);
  List_fill_array((void **) array3,transcripts3);
  qsort(array5,ntranscripts5,sizeof(T),Transcript_velocity_cmp);
  qsort(array3,ntranscripts3,sizeof(T),Transcript_velocity_cmp);

  
  n5 = n3 = 0;
  if (search_velocity <= V_SPLICED) {
    n5 += nspliced5; n3 += nspliced3;
  }
  if (search_velocity <= V_AMBIGUOUS) {
    n5 += nambiguous5; n3 += nambiguous3;
  }
  if (search_velocity <= V_EXTENDED) {
    n5 += nextended5; n3 += nextended3;
  }
  if (search_velocity <= V_UNSPLICED) {
    n5 += nunspliced5; n3 += nunspliced3;
  }
  if (search_velocity <= V_INVALID) {
    n5 = ntranscripts5; n3 = ntranscripts3;
  }


  /* Re-sort beginning of arrays ignoring velocity */
  qsort(array5,n5,sizeof(T),Transcript_cmp);
  qsort(array3,n3,sizeof(T),Transcript_cmp);

  i = k = 0;
  while (i < n5 && k < n3) {
    trnum5 = array5[i]->num;
    j = i+1;
    while (j < n5 && array5[j]->num == trnum5) {
      j++;
    }

    trnum3 = array3[k]->num;
    l = k+1;
    while (l < n3 && array3[l]->num == trnum3) {
      l++;
    }
    
    if (trnum5 < trnum3) {
      i = j;
      
    } else if (trnum3 < trnum5) {
      k = l;
      
    } else {
      besta = bestb = -1;
      best_absdifflength = 0;
      
      for (a = i; a < j; a++) {
	start5 = array5[a]->start;
	end5 = array5[a]->end;
	
	if (desired_genestrand != 0 && array5[a]->transcript_genestrand != desired_genestrand) {
	  /* Not the desired genestrand */
	  debug2(printf("Incorrect genestrand for trnum %d: start5 %d, end3 %d\n",trnum5,start5,end3));

	} else if (start5 < end5) {
	  /* Plus on transcript */
	  for (b = k; b < l; b++) {
	    end3 = array3[b]->end;
	    debug2(printf("(5) Checking for concordance on trnum %d: start5 %d, end3 %d\n",trnum5,start5,end3));
	    if (start5 < end3 && end3 <= start5 + pairmax_transcriptome) {
	      if ((insertlength = end3 - start5) < expected_pairlength) {
		absdifflength = expected_pairlength - insertlength;
	      } else {
		absdifflength = insertlength - expected_pairlength;
	      }
	      if (besta < 0 || absdifflength < best_absdifflength) {
		besta = a;
		bestb = b;
		best_absdifflength = absdifflength;
	      }
	    }
	  }
	  
	} else {
	  /* Minus on transcript */
	  for (b = k; b < l; b++) {
	    end3 = array3[b]->end;
	    debug2(printf("(6) Checking for concordance on trnum %d: end3 %d, start5 %d\n",trnum5,end3,start5));
	    if (end3 < start5 && start5 <= end3 + pairmax_transcriptome) {
	      if ((insertlength = start5 - end3) < expected_pairlength) {
		absdifflength = expected_pairlength - insertlength;
	      } else {
		absdifflength = insertlength - expected_pairlength;
	      }
	      if (besta < 0 || absdifflength < best_absdifflength) {
		besta = a;
		bestb = b;
		best_absdifflength = absdifflength;
	      }
	    }
	  }
	}
      }
      
      if (besta >= 0) {
	if (velocity_accept_cmp(array5[besta]->velocity,array3[bestb]->velocity) > 0) {
	  velocity = array5[besta]->velocity = array3[bestb]->velocity;
	} else {
	  velocity = array3[bestb]->velocity = array5[besta]->velocity;
	}

#if 0
	if (array5[besta]->velocity == V_INVALID || array3[bestb]->velocity == V_INVALID) {
	  velocity = array5[besta]->velocity = array3[bestb]->velocity = V_INVALID;

	} else if (array5[besta]->velocity == V_UNSPLICED || array3[bestb]->velocity == V_UNSPLICED) {
	  velocity = array5[besta]->velocity = array3[bestb]->velocity = V_UNSPLICED;

	} else if (array5[besta]->velocity == V_EXTENDED || array3[bestb]->velocity == V_EXTENDED) {
	  velocity = array5[besta]->velocity = array3[bestb]->velocity = V_EXTENDED;

	} else if (array5[besta]->velocity == V_SPLICED || array3[bestb]->velocity == V_SPLICED) {
	  velocity = array5[besta]->velocity = array3[bestb]->velocity = V_SPLICED;

	} else {
	  /* Both must be V_AMBIGUOUS */
	  velocity = /* array5[besta]->velocity = array3[bestb]->velocity = */ V_AMBIGUOUS;
	}
#endif

	debug2(printf("Comparing velocity %d with accept velocity %d\n",velocity,accept_velocity));
	if (velocity == accept_velocity) {
	  if (first_read_p == true) {
	    concordant = Listpool_push(concordant,listpool,(void *) array5[besta]);
	  } else {
	    concordant = Listpool_push(concordant,listpool,(void *) array3[bestb]);
	  }
	}
      }
      
      i = j;
      k = l;
    }
  }

  FREE(array3);
  FREE(array5);

  return List_reverse(concordant);
}


/* Prints best class using velocity.  Returns genestrand */
List_T
Transcript_get_singleend (int *transcript_genestrand, List_T transcripts,
			  Listpool_T listpool) {
  List_T singleend = NULL, p;
  T *array, transcript;
  int n, i, nspliced, nambiguous, nextended, nunspliced, ntranscripts;
  bool conflictp = false;

  *transcript_genestrand = 0;

  if (transcripts != NULL) {
    nspliced = nambiguous = nextended = nunspliced = ntranscripts = 0;
    for (p = transcripts; p != NULL; p = List_next(p)) {
      transcript = (T) List_head(p);
      if (transcript->velocity == V_SPLICED) {
	nspliced++;
      } else if (transcript->velocity == V_AMBIGUOUS) {
	nambiguous++;
      } else if (transcript->velocity == V_EXTENDED) {
	nextended++;
      } else if (transcript->velocity == V_UNSPLICED) {
	nunspliced++;
      }
      ntranscripts++;
    }
    
    array = (T *) MALLOC(ntranscripts*sizeof(T));
    List_fill_array((void **) array,transcripts);
    qsort(array,ntranscripts,sizeof(T),Transcript_velocity_cmp);

    if (nspliced > 0) {
      n = nspliced;
    } else if (nambiguous > 0) {
      n = nambiguous;
    } else if (nextended > 0) {
      n = nextended;
    } else if (nunspliced > 0) {
      n = nunspliced;
    } else {
      n = ntranscripts;
    }
      
    singleend = Listpool_push(singleend,listpool,(void *) array[0]);
    *transcript_genestrand = array[0]->transcript_genestrand;
    
    for (i = 1; i < n; i++) {
      transcript = array[i];
      singleend = Listpool_push(singleend,listpool,(void *) transcript);

      if (transcript->transcript_genestrand > 0 && *transcript_genestrand < 0) {
	conflictp = true;
      } else if (transcript->transcript_genestrand < 0 && *transcript_genestrand > 0) {
	conflictp = true;
      } else {
	*transcript_genestrand = transcript->transcript_genestrand;
      }
    }

    FREE(array);
  }

  if (conflictp == true) {
    *transcript_genestrand = 0;
  }

  return List_reverse(singleend);
}


static void
tokens_print_fwd (Filestring_T fp, List_T tokens) {
  List_T p;
  char *token;

  if (tokens != NULL) {
    token = (char *) List_head(tokens);
    FPRINTF(fp,":%s",token);
  
    for (p = List_next(tokens); p != NULL; p = List_next(p)) {
      token = (char *) List_head(p);
      FPRINTF(fp,"|%s",token);
    }
  }

  return;
}


static void
tokens_print_rev (Filestring_T fp, List_T tokens) {
  List_T p;
  char *token, tmp;
  int length;

  if (tokens != NULL) {
    tokens = List_reverse(tokens);

    token = (char *) List_head(tokens);
    length = strlen(token);
    tmp = token[0];
    token[0] = token[length-1];
    token[length-1] = tmp;
    FPRINTF(fp,":%s",token);
  
    for (p = List_next(tokens); p != NULL; p = List_next(p)) {
      token = (char *) List_head(p);
      
      length = strlen(token);
      tmp = token[0];
      token[0] = token[length-1];
      token[length-1] = tmp;
      FPRINTF(fp,"|%s",token);
    }

    tokens = List_reverse(tokens);
  }

  return;
}


static int
print_one (Filestring_T fp, T this, Univ_IIT_T transcript_iit, bool invertp) {
  char *label, velocity_char;
  bool allocp;

  label = Univ_IIT_label(transcript_iit,this->num,&allocp);

  switch (this->velocity) {
  case V_INVALID: velocity_char = 'X'; break;
  case V_UNSPLICED: velocity_char = 'U'; break;
  case V_EXTENDED: velocity_char = 'E'; break;
  case V_AMBIGUOUS: velocity_char = 'A'; break;
  case V_SPLICED: velocity_char = 'S'; break;
  default: abort();
  }

  if (output_type == STD_OUTPUT) {
    /* Follow query */
    if (invertp == false) {
      if (this->start < this->end) {
	FPRINTF(fp,"+%s:%d..%d:%c",label,
		this->start < 0 ? this->start : this->start + 1,
		this->end,
		velocity_char);
	tokens_print_fwd(fp,this->tokens);
      } else {
	FPRINTF(fp,"-%s:%d..%d:%c",label,
		this->start,
		this->end < 0 ? this->end : this->end + 1,
		velocity_char);
	tokens_print_rev(fp,this->tokens);
      }
    } else {
      if (this->start < this->end) {
	FPRINTF(fp,"-%s:%d..%d:%c",label,
		this->end,
		this->start < 0 ? this->start : this->start + 1,
		velocity_char);
	tokens_print_rev(fp,this->tokens);
      } else {
	FPRINTF(fp,"+%s:%d..%d:%c",label,
		this->end < 0 ? this->end : this->end + 1,
		this->start,
		velocity_char);
	tokens_print_fwd(fp,this->tokens);
      }
    }

  } else if (this->transcript_genestrand > 0) {
    if (this->start < this->end) {
      FPRINTF(fp,"+%s:%d..%d:%c",label,
	      this->start < 0 ? this->start : this->start + 1,
	      this->end,
	      velocity_char);
    } else {
      FPRINTF(fp,"+%s:%d..%d:%c",label,
	      this->end < 0 ? this->end : this->end + 1,
	      this->start,
	      velocity_char);
    }
    tokens_print_fwd(fp,this->tokens);

  } else {
    if (this->start < this->end) {
      FPRINTF(fp,"-%s:%d..%d:%c",label,
	      this->end,
	      this->start < 0 ? this->start : this->start + 1,
	      velocity_char);
    } else {
      FPRINTF(fp,"-%s:%d..%d:%c",label,
	      this->start,
	      this->end < 0 ? this->end : this->end + 1,
	      velocity_char);
    }
    tokens_print_rev(fp,this->tokens);
  }


  if (allocp) {
    FREE(label);
  }

  return this->transcript_genestrand;
}


void
Transcript_print_list (Filestring_T fp, List_T transcripts,
		       Univ_IIT_T transcript_iit, bool invertp, char *header) {
  T transcript;
  List_T p;

  if (transcripts != NULL) {
    transcript = List_head(transcripts);
    FPRINTF(fp,"%s",header);
    print_one(fp,transcript,transcript_iit,invertp);

    for (p = List_next(transcripts); p != NULL; p = List_next(p)) {
      transcript = List_head(p);
      PUTC(',',fp);
      print_one(fp,transcript,transcript_iit,invertp);
    }
  }

  return;
}


#if 0
Transcript_velocity_T
Transcript_list_velocity (List_T transcripts) {
  Transcript_velocity_T velocity = V_INVALID;
  List_T p;
  T transcript;
  
  for (p = transcripts; p != NULL; p = List_next(p)) {
    transcript = (T) List_head(p);
    if (transcript->velocity > velocity) {
      tokens_print_stdout(transcript->tokens);
      printf("\nVelocity is now %d\n",transcript->velocity);
      velocity = transcript->velocity;
    }
  }

  return velocity;
}
#endif


